<?php

namespace LakeDrops\Drupal8Theme\SASS;

use Composer\Installer\InstallationManager;
use Composer\Package\Package;
use Exception;
use LakeDrops\Component\Composer\BaseHandler;
use MatthiasMullie\PathConverter\Converter;
use RuntimeException;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Finder\Finder;

/**
 * Class Handler.
 *
 * @package LakeDrops\Drupal8Theme\SASS
 */
class Handler extends BaseHandler {

  /**
   * Whether to update (TRUE) or initially installing (FALSE).
   *
   * @var bool
   */
  protected $update = FALSE;

  /**
   * Whether to reset the theme.
   *
   * @var bool
   */
  protected $reset = FALSE;

  /**
   * Whether to overwrite existing files.
   *
   * @var bool
   */
  protected $overwrite = FALSE;

  /**
   * {@inheritdoc}
   */
  public function configId(): string {
    return 'theme-d8-sass';
  }

  /**
   * {@inheritdoc}
   */
  protected function configDefault(): array {
    return [
      'base_theme' => [
        'name' => 'bartik',
        'import' => [],
      ],
      'bower_assets' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function postInit(): void {

    // If we don't have a project name by now, lets get some magic at work.
    $projectName = $this->config->readValue('project_name');
    if ($projectName === NULL) {
      $defaultName = str_replace([' ', '-', '_', '.'], '', basename(getcwd())) . '_theme';
      $projectName = $this->env->receive('name', 'Name of custom theme', $defaultName);
      $this->config->setValue('project_name', $projectName);
    }

    $fs = new Filesystem();

    $dependencies = [];
    $jsFiles = [];
    $to = $this->buildInstallationPath($projectName);
    $bowerAssets = $this->config->readValue('bower_assets') ?? [];
    foreach ($bowerAssets as $key => $asset) {
      $package = $this->getPackage('bower-asset/' . $key);
      $version = $package ? $package->getVersion() : ($asset['version'] ?? 'latest');
      if ($package) {
        // Installed globally as Composer package.
        $from = $this->getInstallationManager()->getInstallPath($package);
        if (!$fs->exists($from)) {
          // Global package is missing.
          continue;
        }
      }
      else {
        // Will be installed locally via Bower.
        $from = implode(DIRECTORY_SEPARATOR, [$to, 'bower_components', $key]);
        $dependencies[] = '"' . $key . '": "' . $version . '"';
      }
      $converter = new Converter($from, $to);
      foreach (['fonts', 'sass', 'js'] as $type) {
        if (isset($asset[$type])) {
          $bowerAssets[$key][$type]['relative'] = $converter->convert($asset[$type]['src']);
          if ($type === 'js') {
            $jsFiles[$asset['name']] = [];
            $finder = new Finder();
            $finder->ignoreDotFiles(FALSE);
            $finder->files()->in($from . DIRECTORY_SEPARATOR . $asset[$type]['src']);
            foreach ($finder as $file) {
              $jsFiles[$asset['name']][] = $file->getRelativePathname();
            }
          }
          if ($this->reset && in_array($type, ['fonts', 'js'])) {
            $path = implode(DIRECTORY_SEPARATOR, [$to, $type, $asset['name']]);
            if ($fs->exists($path)) {
              $fs->remove($path);
            }
          }
        }
      }
    }
    $bowerAssets['dependencies'] = implode(",\n    ", $dependencies);
    $bowerAssets['jsfiles'] = $jsFiles;
    $this->config->setValue('bower_assets', $bowerAssets);
  }

  /**
   * Set whether to update (TRUE) or initially installing (FALSE).
   *
   * @param bool $flag
   *   The flag.
   *
   * @return $this
   */
  public function setUpdate(bool $flag): self {
    $this->update = $flag;
    return $this;
  }

  /**
   * Set whether to reset (TRUE) the theme or not (FALSE).
   *
   * @param bool $flag
   *   The flag.
   *
   * @return $this
   */
  public function setReset(bool $flag): self {
    $this->reset = $flag;
    return $this;
  }

  /**
   * Set whether to overwrite (TRUE) existing files or not (FALSE).
   *
   * @param bool $flag
   *   The flag.
   *
   * @return $this
   */
  public function setOverwrite(bool $flag): self {
    $this->overwrite = $flag;
    return $this;
  }

  /**
   * Get the composer installation manager.
   *
   * @return \Composer\Installer\InstallationManager
   *   The installation manager.
   */
  protected function getInstallationManager(): InstallationManager {
    static $installationManager;

    if (!isset($installationManager)) {
      $installationManager = $this->composer->getInstallationManager();
    }
    return $installationManager;
  }

  /**
   * Get the installation path for a package by its name.
   *
   * @param string $name
   *   The name of the package.
   *
   * @return string
   *   The installation path for the named package.
   */
  protected function buildInstallationPath(string $name): string {
    $package = new Package($name, '1.0.0', '1.0.0');
    $package->setType('drupal-custom-theme');
    return $this->getInstallationManager()
      ->getInstallPath($package);
  }

  /**
   * Get the installation path for the theme of this package.
   *
   * @return string
   *   The installation path of this theme.
   */
  protected function getInstallationPath(): string {
    static $installationPath;

    if (!isset($installationPath)) {
      $installationPath = $this->buildInstallationPath($this->config->readValue('project_name'));
    }
    return $installationPath;
  }

  /**
   * Get the list of custom files.
   *
   * @return array
   *   The file list.
   */
  protected function getCustomFiles(): array {
    return [
      'config/install/{{ project_name }}.settings.yml',
      'config/schema/{{ project_name }}.schema.yml',
      'js/{{ project_name }}.js',
      'sass/_default-variables.scss',
      'sass/style.scss',
      'sass/environment/development/_local-settings.scss',
      'sass/environment/production/_local-settings.scss',
      '.gitignore',
      'favicon.ico',
      'logo.svg',
      'screenshot.png',
      '{{ project_name }}.info.yml',
      '{{ project_name }}.libraries.yml',
      '{{ project_name }}.theme',
    ];
  }

  /**
   * Get a list of ignored files.
   *
   * @return array
   *   The file list.
   */
  protected function getIgnoredFiles(): array {
    return array_merge((empty($this->config->readValue(['bower_assets', 'dependencies'])) ? ['bower.json'] : []), [
      'templates/README.md',
      'README.md',
      '{{ project_name }}.starterkit.yml',
      '.gitignore',
    ]);
  }

  /**
   * Get a list of binary files.
   *
   * @return array
   *   The file list.
   */
  protected function getBinaryFiles(): array {
    return [
      'favicon.ico',
      'screenshot.png',
    ];
  }

  /**
   * Some file name tweaks.
   *
   * @param string $string
   *   The string that needs to be tweaked.
   *
   * @return string
   *   The tweaked string.
   */
  protected function tweaks(string $string): string {
    $string = str_replace('THEMENAME', '{{ project_name }}', $string);

    $baseTheme = $this->config->readValue('base_theme');
    if (isset($baseTheme['sasspath']) && strpos($string, $baseTheme['sasspath']) === 0) {
      $string = 'sass' . substr($string, strlen($baseTheme['sasspath']));
    }

    return $string;
  }

  /**
   * Twig process a given string and return the result.
   *
   * @param string $string
   *   The string that needs to be processed.
   *
   * @return string
   *   The processed string.
   */
  protected function twig(string $string): string {
    if (empty($string)) {
      return '';
    }

    return str_replace('&quot;', '"', $this->config->render('fixed', $this->tweaks($string)));
  }

  /**
   * Wrapper for npm command.
   *
   * @param string $command
   *   NPM command name, arguments and/or options.
   *
   * @throws \RuntimeException
   */
  protected function npm(string $command): void {
    passthru(escapeshellcmd('npm ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new RuntimeException('NPM returned a non-zero exit code');
    }
  }

  /**
   * Wrapper for bower command.
   *
   * @param string $command
   *   Bower command name, arguments and/or options.
   *
   * @throws \RuntimeException
   */
  protected function bower(string $command): void {
    passthru(escapeshellcmd('bower --allow-root ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new RuntimeException('Bower returned a non-zero exit code');
    }
  }

  /**
   * Wrapper for gulp command.
   *
   * @param string $command
   *   Gulp command name, arguments and/or options.
   *
   * @throws \RuntimeException
   */
  protected function gulp(string $command): void {
    passthru(escapeshellcmd('node_modules/.bin/gulp ' . $command), $exit_code);
    if ($exit_code !== 0) {
      throw new RuntimeException('Gulp returned a non-zero exit code');
    }
  }

  /**
   * Recursively copy files from $source to $destination.
   *
   * @param \Symfony\Component\Filesystem\FileSystem $fs
   *   The file system object.
   * @param string $source
   *   The source directory.
   * @param string $destination
   *   The destination directory.
   *
   * @return array
   *   List of processed file names.
   */
  protected function copyFiles(FileSystem $fs, string $source, string $destination): array {
    $files = [];

    // Make sure all directories do exist.
    $finder = new Finder();
    $finder->ignoreDotFiles(FALSE);
    $finder->directories()->in($source);
    foreach ($finder as $file) {
      $path = $destination . DIRECTORY_SEPARATOR . $this->twig($file->getRelativePathname());
      if (!$fs->exists($path)) {
        $fs->mkdir($path, 0775);
      }
    }

    $finder = new Finder();
    $finder->ignoreDotFiles(FALSE);
    $finder->files()->in($source);
    foreach ($finder as $file) {

      // Do not copy .gitkeep files.
      if ($file->getFilename() === '.gitkeep') {
        continue;
      }

      // Determine relative filename of destination.
      $relFilename = $file->getRelativePathname();
      if ($file->getExtension() === 'twig') {
        $relFilename = substr($relFilename, 0, strrpos($relFilename, '.twig'));
      }
      $tweakedRelFilename = $this->tweaks($relFilename);
      $filename = $destination . DIRECTORY_SEPARATOR . $this->twig($relFilename);

      // Do not copy ignored files.
      if (in_array($tweakedRelFilename, $this->getIgnoredFiles(), true)) {
        continue;
      }
      // Do not update custom files.
      if (!$this->overwrite && $fs->exists($filename) && in_array($tweakedRelFilename, $this->getCustomFiles(), true)) {
        continue;
      }

      // Now create destination.
      if (in_array($tweakedRelFilename, $this->getBinaryFiles(), true)) {
        $fs->copy($file->getRealPath(), $filename);
      }
      else {
        file_put_contents($filename, $this->twig(file_get_contents($file->getRealPath())));
      }
      $files[] = $tweakedRelFilename;
    }

    return $files;
  }

  /**
   * Execute the theme creation or update.
   *
   * @throws Exception
   */
  public function updateTheme(): void {
    $this->init();

    $fs = new Filesystem();
    $destination = $this->getInstallationPath();

    $pluginRoot = $this->getInstallationManager()
      ->getInstallPath($this->getPackage('lakedrops/theme-d8-sass'));

    // Copy all files.
    $this->copyFiles($fs, $pluginRoot . '/templates', $destination);

    // Handle optional base theme.
    $baseTheme = $this->config->readValue('base_theme');
    if (isset($baseTheme['package'])) {
      $baseThemePackage = $this->getPackage($baseTheme['package']);
      if ($baseThemePackage && isset($baseTheme['starterkit'])) {
        $starterkitRoot = $this->getInstallationManager()->getInstallPath($baseThemePackage) . DIRECTORY_SEPARATOR . $baseTheme['starterkit'];
        $this->config->setValue('base_theme_files', $this->copyFiles($fs, $starterkitRoot, $destination));
      }
    }

    // Remove bower stuff if that's available globally.
    if ($this->config->readValue(['bower_assets', 'dependencies']) === NULL) {
      if ($fs->exists($destination . DIRECTORY_SEPARATOR . 'bower.json')) {
        $fs->remove($destination . DIRECTORY_SEPARATOR . 'bower.json');
      }
      if ($fs->exists($destination . DIRECTORY_SEPARATOR . 'bower_components')) {
        $fs->remove($destination . DIRECTORY_SEPARATOR . 'bower_components');
      }
    }

    // Finally write .gitignore.
    file_put_contents($destination . '/.gitignore', $this->twig(file_get_contents($pluginRoot . '/templates/.gitignore.twig')));

    // Prepare post processing.
    $currentDir = getcwd();
    chdir($destination);

    // Handle NPM.
    if ($this->update) {
      $this->npm('update');
    }
    else {
      $this->npm('install --no-shrinkwrap');
    }

    // Handle Bower if not installed globally.
    if ($fs->exists($destination . DIRECTORY_SEPARATOR . 'bower.json') && $this->config->readValue(['bower_assets', 'dependencies']) !== NULL) {
      if ($this->update) {
        $this->bower('update');
      }
      else {
        $this->bower('install');
      }
    }

    // Run Gulp.
    $args = ['default'];
    if (!$this->isDevMode()) {
      $args[] = '--env production';
    }
    $this->gulp(implode(' ', $args));

    // Restore current directory.
    chdir($currentDir);
  }

}
